﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Windows.Forms;
using System.Resources;
using System.Threading;
using System.Reflection;
using System.IO;
using System.Globalization;
using RootPro.RootProCAD;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.UI;

/// <summary>
/// 三斜面積計算アドインプログラム
/// 概要
/// 　敷地の外周を表す線とその外周を三角形に分割する線を選択すると、三斜面積計算を行い面積表を自動的に作図することのできる RootPro CAD のアドインプログラムです。
/// 　図形の座標値が大きい場合は、用紙上に図形を配置してコマンドを実行することで三角形の交点を正確に計算できるようになります。
/// 　(線の長さは mm として扱いますが、結果の長さや面積は m に換算しています。)
/// ※使用許諾
/// ・このアドインを実行した結果についての一切の責任はお使いになられる方にあります。
/// ・このアドインは、プログラムソース自体を改変・流用することが可能です。
/// 　また、このプログラムソースを元に独自のアドインやソフトウェアを作成し、無償・有償を問わず第三者に配布することも認めます。
/// 　ただし、改変・流用する場合は、株式会社ルートプロ(RootPro Corporation)作成の
/// 　当プログラムソースを参考にしていることを第三者にわかるように明記してください。
/// 　
/// 　Copyright @ 2020 RootPro Co.,Ltd. (Japan)
/// </summary>

namespace RCAddInTriangularQuadratureCS
{
    [System.AddIn.AddIn("AppAddIn", Version = "1.0", Publisher = "", Description = "")]
    public partial class AppAddIn
    {
        // アドインプログラム名
        string AddInProgramName = "RootPro CAD";

        List<Shape> rubberBandShapes = null;       // ラバーバンド図形リスト
        List<Shape> rubberBandShapes2 = null;       // ラバーバンド図形リスト

        string TriangularQuadratureCommandName = "RootPro.AddIn.TriangularQuadratureCommand";     //コマンド名 

        int RC_MAX_SHAPES_COUNT = 2400;// 最大図形数

        // エラー番号
        const int RC_ERR_SUCCESS = 1; // エラーなし（成功）
        const int RC_ERR_OVER_SHAPES_COUNT = -10;// エラー：図形数が多すぎます。
        const int RC_ERR_DIFFERENT_CURRENTDRAWING = -11;// エラー：選択した図形にカレントの部分図以外のものが含まれています。
        const int RC_ERR_DIFFERENT_LAYER = -12;// エラー：選択した図形にレイヤが異なるものがあります。
        const int RC_ERR_DIFFERENT_COLOR = -13;// エラー：選択した図形に色番号が異なるものがあります。
        const int RC_ERR_DIFFERENT_LINETYPE = -14;// エラー：選択した図形に線種番号が異なるものがあります。
        const int RC_ERR_DIFFERENT_LINEWIDTH = -15;// エラー：選択した図形に線幅番号が異なるものがあります。
        const int RC_ERR_DONT_CREATE_TRIANGLE = -20; // エラー：三角形が見つかりませんでした。


        /// <summary>
        /// アドイン起動時に呼ばれる処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AppAddIn_Startup(object sender, EventArgs e)
        {
            // 多言語対応リソースのDLLファイルをロード
            AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(ResolveAssembly);

            // コマンドの追加
            CommandManager.AddMacroCommand("RCAddInTriangularQuadratureCS", Properties.Resources.CommandName, MacroCommand);

            Command command = CommandManager.AddCommand(TriangularQuadratureCommandName);
            command.Name = Properties.Resources.CommandName;

            // パラメータの追加
            // 第1パラメーター 「図形」
            CommandParameterGroup commandParameterGroup = command.ParameterGroups.Add(ParameterItemType.ShapeGroup, 0, Properties.Resources.ShapeGroupParameterName);
            ShapeGroupParameterItem parameterItem = (ShapeGroupParameterItem)commandParameterGroup.CurrentItem; // パラメータアイテム
            parameterItem.ShapeTypeFilter = ShapeTypeFilters.Line | ShapeTypeFilters.Polyline;    // 線分、連続線
            parameterItem.Description = Properties.Resources.ShapeGroupParameterDescription;

            // 第2パラメーター「開始番号」
            command.ParameterGroups.Add(ParameterItemType.Integer, 1, Properties.Resources.StartListNumberParameterName);
            IntegerParameterItem integerParameterItem = (IntegerParameterItem)command.GetParameterItemByID(1);
            integerParameterItem.Description = Properties.Resources.StartListNumberParameterDescription;
            integerParameterItem.Value = 1;
            integerParameterItem.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法
            
            // 第3パラメーター「配置点」
            command.ParameterGroups.Add(ParameterItemType.Point, 2, Properties.Resources.TableLocatePointParameterName);
            PointParameterItem pointParameterItem = (PointParameterItem)command.GetParameterItemByID(2);
            pointParameterItem.Description = Properties.Resources.TableLocatePointParameterDescription;

            //開始イベントと終了イベントの追加
            command.Begin += new CommandBeginEventHandler(command_Begin);
            command.End += new EventHandler(command_End);
        }

        /// <summary>
        /// アドイン終了時に呼ばれる処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AppAddIn_Shutdown(object sender, EventArgs e)
        {
            CommandManager.RemoveCommand(TriangularQuadratureCommandName);
            CommandManager.RemoveMacroCommand(MacroCommand); // 追加したコマンド関数
        }

        #region VSTA generated code
        private void InternalStartup()
        {
            this.Startup += new System.EventHandler(AppAddIn_Startup);
            this.Shutdown += new System.EventHandler(AppAddIn_Shutdown);
        }
        #endregion

        /// <summary>
        /// 外部フォルダのアセンブリを動的にロード
        /// 多言語対応リソースのDLLファイルをロード
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        private Assembly ResolveAssembly(object sender, ResolveEventArgs args)
        {
            AssemblyName assemblyName = new AssemblyName(args.Name);

            foreach(Assembly assembly in AppDomain.CurrentDomain.GetAssemblies())
            {
                if(AssemblyName.ReferenceMatchesDefinition(assemblyName, assembly.GetName()))
                    return assembly;
            }

            if(assemblyName.Name.EndsWith(".resources"))
            {
                if(assemblyName.CultureName.Equals(Thread.CurrentThread.CurrentUICulture.Name, StringComparison.OrdinalIgnoreCase))
                {
                    string dllPath = Path.GetDirectoryName(this.AddInFullPath);
                    dllPath = Path.Combine(dllPath, assemblyName.CultureName);
                    dllPath = Path.Combine(dllPath, assemblyName.Name + ".dll");

                    if(File.Exists(dllPath))
                    {
                        try
                        {
                            return Assembly.LoadFrom(dllPath);
                        }
                        catch
                        {
                        }
                    }
                }
            }

            return null;
        }


        /// <summary>
        /// マクロコマンドの実行
        /// </summary>
        private void MacroCommand()
        {
            //コマンドの実行
            CommandManager.ExecuteCommand(TriangularQuadratureCommandName);
        }

        /// <summary>
        /// コマンドの開始時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_Begin(Object sender, CommandBeginEventArgs e)
        {
            Command command = (Command)sender;

            //パラメータ変更イベントの追加
            command.ParameterChanged += new CommandParameterChangedEventHandler(Command_ParameterChanged);

            //マウス移動イベントの追加
            command.MouseMove += new CommandMouseMoveEventHandler(Command_MouseMove);

            // 図形の既定値設定変更イベントハンドラ追加
            ShapeSettingsChanged += new ShapeSettingsEventHandler(App_ShapeSettingsChanged);
            // カレント部分図変更イベントハンドラ追加
            CurrentDrawingChanged += new CurrentDrawingChangedEventHandler(App_CurrentDrawingChanged);
        }

        /// <summary>
        /// コマンドの終了時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_End(Object sender, EventArgs e)
        {
            Command command = (Command)sender;

            //パラメータ変更イベントの削除
            command.ParameterChanged -= new CommandParameterChangedEventHandler(Command_ParameterChanged);

            //マウス移動イベントの削除
            command.MouseMove -= new CommandMouseMoveEventHandler(Command_MouseMove);

            // 図形の既定値設定変更イベントの削除
            ShapeSettingsChanged -= new ShapeSettingsEventHandler(App_ShapeSettingsChanged);
            // カレント部分図変更イベントの削除
            CurrentDrawingChanged -= new CurrentDrawingChangedEventHandler(App_CurrentDrawingChanged);

            //ラバーバンドクリア
            ClearRubberBand();
        }

        /// <summary>
        /// 図形の既定値設定の変更イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_ShapeSettingsChanged(object sender, ShapeSettingsEventArgs e)
        {
            //ラバーバンドクリア
            ClearRubberBand();
        }

        /// <summary>
        /// カレント部分図の変更イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void App_CurrentDrawingChanged(object sender, CurrentDrawingChangedEventArgs e)
        {
            //ラバーバンドクリア
            ClearRubberBand();
        }

        /// <summary>
        /// ラバーバンドクリア
        /// </summary>
        private void ClearRubberBand()
        {
            if(rubberBandShapes != null)
            {
                rubberBandShapes.Clear();
                rubberBandShapes = null;
            }

            if(rubberBandShapes2 != null)
            {
                rubberBandShapes2.Clear();
                rubberBandShapes2 = null;
            }
        }
        
        /// <summary>
        /// コマンド実行中にマウスが移動イベントハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Command_MouseMove(Object sender, CommandMouseMoveEventArgs e)
        {
            Command command = (Command)sender;
            // Command.GetTemporaryParameterItemByID を使用してコマンドの一時的なパラメータアイテムを取得します。
            // 一時的なパラメータアイテムからは、すでにパラメータの入力が確定している場合は確定している入力値が、
            // 入力が確定していない場合はマウスの現在位置などの一時的な値が取得できます。

            // 選択した図形の取得
            ShapeGroupParameterItem shapeGroupParameterItem = (ShapeGroupParameterItem)command.GetTemporaryParameterItemByID(0);
            if(shapeGroupParameterItem == null)
                return;

            if(shapeGroupParameterItem.IsEmpty)
                return;

            SelectedShapeCollection selectedShapeCollection = shapeGroupParameterItem.SelectedShapes;

            // 開始番号
            IntegerParameterItem integerParameterItem = (IntegerParameterItem)command.GetTemporaryParameterItemByID(1);
            if(integerParameterItem == null)
                return;

            if(integerParameterItem.IsEmpty)
                return;

            int startNumber = integerParameterItem.Value;

            // 表の配置点
            PointParameterItem pointParameterItem = (PointParameterItem)command.GetTemporaryParameterItemByID(2);
            if(pointParameterItem == null)
                return;

            if(pointParameterItem.IsEmpty)
                return;

            Point2d locatePoint = pointParameterItem.Point;

            //ドキュメントの取得
            Document doc = ActiveDocument;
            //部分図の取得
            Drawing drawing = doc.CurrentDrawing;

            // ラバーバンド用の図形リスト
            Point2d tempOrgPoint = new Point2d(0.0, 0.0);
            if(rubberBandShapes == null || rubberBandShapes2 == null)
            {
                //ラバーバンド表示を削除する
                ClearRubberBand();

                // 三斜求積処理開始 図形作成
                List<Shape> triclinicTrianglesShapes = null;  // 三斜三角形の図形
                List<Shape> triclinicTrianglesTableShapes = null;// 三斜求積表の図形
                int stat = CreateCommandShapes(selectedShapeCollection, startNumber, tempOrgPoint, out triclinicTrianglesShapes, out triclinicTrianglesTableShapes);
                if(stat < 0)
                {
                    return;
                }

                 if(rubberBandShapes == null)
                    rubberBandShapes = new List<Shape>();

                 rubberBandShapes.AddRange(triclinicTrianglesShapes);

                if(rubberBandShapes2 == null)
                    rubberBandShapes2 = new List<Shape>();

                 rubberBandShapes2.AddRange(triclinicTrianglesTableShapes);
            }

            // 表示させるラバーバンド図形
            List<Shape> thisRubberBandShapes = new List<Shape>();
            thisRubberBandShapes.AddRange(rubberBandShapes);// 三斜三角形の図形
            // 三斜求積表の図形
            foreach(Shape shape in rubberBandShapes2)
            {
                // マウスの位置に移動
                Shape shape2 = shape.Clone();
                shape2.Transform(tempOrgPoint, 1.0, 1.0, 0.0, locatePoint);
                thisRubberBandShapes.Add(shape2);
            }

            try
            {
                // ラバーバンド描画
                command.DrawShapeRubberBand(thisRubberBandShapes.ToArray());
            }
            catch
            {
                throw;
            }

        }


        /// <summary>
        /// コマンドパラメータ変更時の処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Command_ParameterChanged(Object sender, CommandParameterChangedEventArgs e)
        {
            //ラバーバンド表示を削除する
            ClearRubberBand();

            //ドキュメントの取得
            Document doc = ActiveDocument;
            Drawing drawing = doc.CurrentDrawing;

            Command command = (Command)sender;
            int id = e.ParameterID;// 変更のあったパラメータID
            ParameterItem parameterItem = command.GetParameterItemByID(id); // パラメータアイテム

            //図形の取得
            ShapeGroupParameterItem shapeGroupParameterItem = (ShapeGroupParameterItem)command.GetParameterItemByID(0);
            if(shapeGroupParameterItem == null)
                return;

            if(shapeGroupParameterItem.IsEmpty)
                return;

            SelectedShapeCollection selectedShapeCollection = shapeGroupParameterItem.SelectedShapes;

            // 実行可能かチェック
            int stat = CheckCanDoCommand(selectedShapeCollection);
            if(stat < 0)
            {
                ShowErrMessage(stat);

                e.CommandState = CommandState.CommandError;
                return;
            }

            // 開始番号
            IntegerParameterItem integerParameterItem = (IntegerParameterItem)command.GetParameterItemByID(1);
            if(integerParameterItem == null)
                return;

            if(integerParameterItem.IsEmpty)
                return;

            int startNumber = integerParameterItem.Value;

            // 表の配置点
            PointParameterItem pointParameterItem = (PointParameterItem)command.GetParameterItemByID(2);
            if(pointParameterItem == null)
                return;

            if(pointParameterItem.IsEmpty)
                return;

            Point2d locatePoint = pointParameterItem.Point;

            // 時間計測
            Stopwatch sw = new Stopwatch();
            sw.Start();

            // 三斜求積処理開始 図形作成
            List<Shape> triclinicTrianglesShapes = null;  // 三斜三角形の図形
            List<Shape> triclinicTrianglesTableShapes = null;// 三斜求積表の図形
            stat = CreateCommandShapes(selectedShapeCollection, startNumber, locatePoint, out triclinicTrianglesShapes, out triclinicTrianglesTableShapes);
            if(stat < 0)
            {
                ShowErrMessage(stat);

                e.CommandState = CommandState.ParameterError;
                return;
            }

            // UNDO 可能な操作を開始
            doc.UndoManager.BeginUndoUnit();

            // 図面上にプロット
            // 三斜三角形
            foreach(Shape shape in triclinicTrianglesShapes)
            {
                // 描画図形追加
                drawing.Shapes.Add(shape);
            }

            // 三斜求積表
            foreach(Shape shape in triclinicTrianglesTableShapes)
            {
                // 描画図形追加
                drawing.Shapes.Add(shape);
            }

            // UNDO 可能な操作を終了
            doc.UndoManager.EndUndoUnit();

            //e.CommandState = CommandState.Execute;
            e.CommandState = CommandState.End;
            
            sw.Stop();
        }

        /// <summary>
        /// 実行可能な図形かチェック
        /// カレント部分図、レイヤ、色、線種、線幅について同じものだけ有効とする
        /// </summary>
        /// <param name="selectedShapeCollection"></param>
        /// <returns>戻り値：0未満はエラー</returns>
        private int CheckCanDoCommand(SelectedShapeCollection selectedShapeCollection)
        {
            //ドキュメントの取得
            Document doc = ActiveDocument;
            Drawing currentDrawing = doc.CurrentDrawing;

            // 選択図形数チェック
            int shapesCount = selectedShapeCollection.Count;
            // エラー：図形数が多すぎます。
            if(shapesCount > RC_MAX_SHAPES_COUNT)
                return RC_ERR_OVER_SHAPES_COUNT;

            int layerId = 0;
            int colorNumber = 0;
            int linetypeNumber = 0;
            int linewidthNumber = 0;
            for(int i = 0; i < shapesCount; i++)
            {
                Shape shape = selectedShapeCollection[i].Shape;

                // 部分図チェック
                // エラー：選択した図形にカレントの部分図以外のものが含まれています。
                if(currentDrawing.ID != shape.Drawing.ID)
                    return RC_ERR_DIFFERENT_CURRENTDRAWING;

                // レイヤ、色、線種、線幅が同じものだけかチェック
                if(i == 0)
                {
                    layerId = shape.LayerID;
                    colorNumber = ((BasicShape)shape).ColorNumber;
                    linetypeNumber = ((BasicShape)shape).LinetypeNumber;
                    linewidthNumber = ((BasicShape)shape).LinewidthNumber;
                }
                else
                {
                    // エラー：選択した図形にレイヤが異なるものがあります。
                    if(layerId != shape.LayerID)
                        return RC_ERR_DIFFERENT_LAYER;

                    // エラー：選択した図形に色番号が異なるものがあります。
                    if(colorNumber != ((BasicShape)shape).ColorNumber)
                        return RC_ERR_DIFFERENT_COLOR;

                    // エラー：選択した図形に線種番号が異なるものがあります。
                    if(linetypeNumber != ((BasicShape)shape).LinetypeNumber)
                        return RC_ERR_DIFFERENT_LINETYPE;

                    // エラー：選択した図形に線幅番号が異なるものがあります。
                    if(linewidthNumber != ((BasicShape)shape).LinewidthNumber)
                        return RC_ERR_DIFFERENT_LINEWIDTH;
                }
            }

            return RC_ERR_SUCCESS;
        }

        /// <summary>
        /// エラーメッセージの表示
        /// </summary>
        /// <param name="errorNumber">エラー番号</param>
        private void ShowErrMessage(int errorNumber)
        {
            string strErr = "";
            switch(errorNumber)
            {
                // 図形数が多すぎます。
                case RC_ERR_OVER_SHAPES_COUNT:
                    strErr = Properties.Resources.ErrOverShapesCount;
                    break;
                // 選択した図形にカレントの部分図以外のものが含まれています。
                case RC_ERR_DIFFERENT_CURRENTDRAWING:
                    strErr = Properties.Resources.ErrDifferentCurrentdrawing;
                    break;
                // 選択した図形にレイヤが異なるものがあります。
                case RC_ERR_DIFFERENT_LAYER:
                    strErr = Properties.Resources.ErrDifferentLayer;
                    break;
                // 選択した図形に色番号が異なるものがあります。
                case RC_ERR_DIFFERENT_COLOR:
                    strErr = Properties.Resources.ErrDifferentColor;
                    break;
                // 選択した図形に線種番号が異なるものがあります。
                case RC_ERR_DIFFERENT_LINETYPE:
                    strErr = Properties.Resources.ErrDifferentLinetype;
                    break;
                // 選択した図形に線幅番号が異なるものがあります。
                case RC_ERR_DIFFERENT_LINEWIDTH:
                    strErr = Properties.Resources.ErrDifferentLinewidth;
                    break;
                // 三角形が見つかりませんでした。
                case RC_ERR_DONT_CREATE_TRIANGLE:
                    strErr = Properties.Resources.ErrDontCreateTriangle;
                    break;

            }

            if(strErr != "")
                MessageBox.Show(strErr, AddInProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);

        }

        /// <summary>
        /// コマンド実行による図形の作成
        /// </summary>
        /// <param name="selectedShapeCollection"></param>
        /// <param name="startNumber"></param>
        /// <param name="locatePoint"></param>
        /// <param name="triclinicTrianglesShapes"></param>
        /// <param name="triclinicTrianglesTableShapes"></param>
        /// <returns></returns>
        private int CreateCommandShapes(SelectedShapeCollection selectedShapeCollection, int startNumber, Point2d locatePoint, out List<Shape> triclinicTrianglesShapes, out List<Shape> triclinicTrianglesTableShapes)
        {
            //ドキュメントの取得
            Document doc = ActiveDocument;
            Drawing drawing = doc.CurrentDrawing;

            triclinicTrianglesShapes = new List<Shape>();// 三斜三角形の図形
            triclinicTrianglesTableShapes = new List<Shape>();// 三斜求積表の図形

            List<SelectedShape> lineShapes = new List<SelectedShape>();
            foreach(SelectedShape selectedShape in selectedShapeCollection)
            {
                Shape shape = selectedShape.Shape;
                // 連続線は線分に
                if(shape is PolylineShape)
                {
                    LineShape[] lineShapes2 = ((PolylineShape)shape).GetSplitLineShapes();
                    foreach(LineShape lineShape in lineShapes2)
                        lineShapes.Add(new SelectedShape(lineShape, selectedShape.Layout));
                }
                else
                {
                    lineShapes.Add(selectedShape);
                }
            }
                
            // 他の線分と端点が一致する線分は除外する
            List<int> delIndexList = new List<int>();
            var lineCount = lineShapes.Count;
            for(int i = 0; i < lineCount - 1; i++)
            {
                for(int j = i + 1; j < lineCount; j++)
                {
                    if(lineShapes[i].Shape is LineShape line1 &&
                        lineShapes[j].Shape is LineShape line2)
                    {
                        // 二つの線分が一致する場合
                        if(IsLineEquals(line1, line2, lineShapes[i].Layout.ScaleX))
                        {
                            // 削除リストに追加する
                            if(!delIndexList.Contains(j))
                                delIndexList.Add(j);
                        }
                    }
                }
            }

            // 降順にソート
            delIndexList.Sort();
            delIndexList.Reverse();

            // 重複した線を削除する
            foreach(var index in delIndexList)
                lineShapes.RemoveAt(index);
            // 線分上に他の線と交差しているものはその交点で分ける
            List<SelectedShape> lineShapes3 = CutCrossLineShapes(lineShapes);

            // 三斜三角形の作成
            List<TriclinicTriangle> triclinicTriangles = CreateTriclinicTriangles(lineShapes3, startNumber);
            if(triclinicTriangles.Count == 0)
            {
                return RC_ERR_DONT_CREATE_TRIANGLE;
            }

            // 作図図形作成
            // 三斜三角形の高さや底辺長さを表す図形作成
            triclinicTrianglesShapes = CreateTriangularQuadratureShapes(drawing, triclinicTriangles);

            // 三斜求積表の図形作成
            triclinicTrianglesTableShapes = CreateTriangularQuadratureTableShapes(drawing, triclinicTriangles, locatePoint);

            return RC_ERR_SUCCESS;
        }
 

    }
}
