﻿using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using RootPro.RootProCAD;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.Geometry;

/// <summary>
/// 三斜面積計算アドインプログラム
/// 概要
/// 　敷地の外周を表す線とその外周を三角形に分割する線を選択すると、三斜面積計算を行い面積表を自動的に作図することのできる RootPro CAD のアドインプログラムです。
/// 　(線の長さは mm として扱いますが、結果の長さや面積は m に換算しています。)
/// ※使用許諾
/// ・このアドインを実行した結果についての一切の責任はお使いになられる方にあります。
/// ・このアドインは、プログラムソース自体を改変・流用することが可能です。
/// 　また、このプログラムソースを元に独自のアドインやソフトウェアを作成し、無償・有償を問わず第三者に配布することも認めます。
/// 　ただし、改変・流用する場合は、株式会社ルートプロ(RootPro Corporation)作成の
/// 　当プログラムソースを参考にしていることを第三者にわかるように明記してください。
/// 　
/// 　Copyright @ 2020 RootPro Co.,Ltd. (Japan)
/// </summary>

namespace RCAddInTriangularQuadratureCS
{
    public partial class AppAddIn
    {
        static private double RC_LengthScale = 0.001;// 実寸をメートルに換算するスケール値 mm -> m

        private int RC_LengthRoundMode = 2;// 長さの丸めモード　0:「切り捨て」1:「切り上げ」2:「四捨五入」 // 2021/01/04 初期値を2に変更
        private int RC_AreaRoundMode = 2;// 面積の丸めモード　0:「切り捨て」1:「切り上げ」2:「四捨五入」   // 2021/01/04 初期値を2に変更
        private int RC_LengthValueDecimals = 2;// 長さ有効数字：小数点以下2位
        private int RC_AreaWValueDecimals = 4;// 倍面積有効数字：小数点以下4位
        private int RC_AreaValueDecimals = 5;// 面積有効数字：小数点以下5位
        private int RC_SiteAreaValueDecimals = 2;// 敷地面積有効数字：小数点以下2位

        // 数値を比較するときに使用する精度
        static double RcPrecisionConfusion = Precision.Confusion;
        static double RcPrecisionAngular = Precision.RadianAngular;
        static double RcPrecisionScale = 1.0E-10;
 
        // 三斜三角形
        private const int RcAddinTriangularQuadratureHightLineColorNumber = -1;  // 三角形の高さの線の色(3:緑、-1:既定値) 
        private const int RcAddinTriangularQuadratureHightLineTypeNumber = 8;   // 三角形の高さの線の線種(8:一点鎖線、-1:既定値)
        private const int RcAddinTriangularQuadratureHightLineWidthNumber = 1;  // 三角形の高さの線の太さ(1:0.13、-1:既定値)
        private const int RcAddinTriangularQuadratureTextColorNumber = -1;// 文字の色(3:緑、-1:既定値) 

        // 三斜求積表
        private const int RcAddinTriangularQuadratureTableLineColorNumber = -1;// 表の線の色(3:緑、-1:既定値) 
        private const int RcAddinTriangularQuadratureTableLineLineWidthNumber = 1;// 表の線の太さ(1:0.13、-1:既定値)
        private const int RcTriangularQuadratureTableWidthTextScale = 30;// 1文字図形の高さに対する三斜求積表の全体の幅倍率
        private const int RcAddinTriangularQuadratureTableTextColorNumber = -1;// 文字の色(3:緑、-1:既定値) 

        /// <summary>
        /// 三斜三角形クラス
        /// </summary>
        public class TriclinicTriangle
        {
             // 番号
            public int triangleNumber { get; private set; }

            // 頂点
            public Point2d vertexPoint { get; }

            // 底辺の始点
            public Point2d baseLineStartPoint { get; }     

            // 底辺の終点
            public Point2d baseLineEndPoint { get; }     

           // 高さ(m)
            public double hight { get; }

            // 頂点から底辺に下した垂直点
            public Point2d verticalPointOnBaseLine { get; }

            // 底辺の長さ(m)
            public double baselineLength { get; }

            // 倍面積(m^2)
            //public double triangleAreaW { get; }// 2021/01/04 変更:表の底辺と高さの値を使った計算結果と同じにするためにここでは求めない。

            // 重心座標
            public Point2d gravityPoint { get; }

            /// <summary>
            /// コンストラクタ―
            /// </summary>
            /// <param name="triangleNumber"></param>
            /// <param name="point1"></param>
            /// <param name="point2"></param>
            /// <param name="point3"></param>
            /// <param name="shapeFactory"></param>
            public TriclinicTriangle(int triangleNumber, Point2d point1, Point2d point2, Point2d point3, ShapeFactory shapeFactory, Geometry geometry)
            {
                hight = 0.0;// 高さ初期値

                if(geometry.IsPointEquals(point1, point2) || geometry.IsPointEquals(point2, point3) || geometry.IsPointEquals(point3, point1))
                    return;

                Debug.Assert(triangleNumber > 0);
                this.triangleNumber = triangleNumber;

                // 3点からなる3辺の内一番長い線を底辺として、その底辺の端点以外の点を頂点とする
                double lineLength12 = Get2PointsDistance(point1, point2);
                double lineLength23 = Get2PointsDistance(point2, point3);
                double lineLength31 = Get2PointsDistance(point3, point1);

                // point1, point2 間が1番長い場合
                if(lineLength12 >= lineLength23 && lineLength12 >= lineLength31)
                {
                    // 頂点
                    vertexPoint = point3;

                    // 底辺の始点
                    baseLineStartPoint = point1;

                    // 底辺の終点
                    baseLineEndPoint = point2;
                }
                // point2, point3 間が1番長い場合
                else if(lineLength23 >= lineLength12 && lineLength23 >= lineLength31)
                {
                    // 頂点
                    vertexPoint = point1;

                    // 底辺の始点
                    baseLineStartPoint = point2;

                    // 底辺の終点
                    baseLineEndPoint = point3;
                }
                else
                {
                    // 頂点
                    vertexPoint = point2;

                    // 底辺の始点
                    baseLineStartPoint = point3;

                    // 底辺の終点
                    baseLineEndPoint = point1;
                }

                // 頂点から底辺に下した垂直点
                LineShape baseLine = shapeFactory.CreateLine(baseLineStartPoint, baseLineEndPoint);
                baselineLength = baseLine.Length  * RC_LengthScale; // 底辺の長さ mm -> m
                if(baselineLength > RcPrecisionConfusion)
                {
                    // 頂点から底辺への垂点
                    PointOnShape pointOnShape = baseLine.GetClosestPoint(vertexPoint);
                    verticalPointOnBaseLine = pointOnShape.Point;

                    // 高さ(m)(頂点が底辺上だと0になる)
                    hight = Get2PointsDistance(vertexPoint, verticalPointOnBaseLine) * RC_LengthScale; // mm -> m

                    // 倍面積(m^2)
                    // triangleAreaW = baselineLength * hight; // * 0.5;誤差を小さくするために/2する前の値を保持 // 2021/01/04 変更:表の底辺と高さの値を使った計算結果と同じにするためにここでは求めない。

                    // 重心
                    if(baseLine.Length > RcPrecisionConfusion)
                    {
                        Point2d centerPointOnBaseline = geometry.CreatePoint((baseLine.StartPoint.X + baseLine.EndPoint.X) * 0.5, (baseLine.StartPoint.Y + baseLine.EndPoint.Y) * 0.5);
                        LineShape tempLine = shapeFactory.CreateLine(vertexPoint, centerPointOnBaseline);
                        gravityPoint = geometry.CreatePoint(vertexPoint.X + tempLine.Length * 2.0 / 3.0, vertexPoint.Y);
                        gravityPoint = geometry.RotatePoint(gravityPoint, vertexPoint, tempLine.Angle);
                    }
                }
            }
        }

        /// <summary>
        /// 2点間の距離取得
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <returns></returns>
        private static double Get2PointsDistance(Point2d point1, Point2d point2)
        {
            return Math.Sqrt((point2.X - point1.X) * (point2.X - point1.X) + (point2.Y - point1.Y) * (point2.Y - point1.Y));
        }

        /// <summary>
        /// 「切り捨て」、「切り上げ」値の取得
        /// </summary>
        /// <param name="value"></param>
        /// <param name="roundUpFlg"></param>
        /// <param name="decimals">小数部桁数</param>
        /// <returns></returns>
        private static double GetRoundValue(double value, bool roundUpFlg, int decimals)
        {
            double roundValue = value;

            // 切り捨ての場合
            if(roundUpFlg == false)
            {
                roundValue = Truncate(value, decimals);//「切り捨て」
            }
            // 切り上げの場合
            else
            {
                roundValue = RoundUp(value, decimals); //「切り上げ」
            }

            return roundValue;
        }


        /// <summary>
        /// 値の指定した桁の小数部を切り捨て丸める
        /// </summary>
        /// <param name="value">値</param>
        /// <param name="decimals">小数部桁数</param>
        /// <returns>丸めた値</returns>
        private static double Truncate(double value, int decimals)
        {
            // 小数部桁数の10の累乗を取得
            double pow = Math.Pow(10, decimals);
            return 0 <= value ? Math.Floor(value * pow) / pow : Math.Ceiling(value * pow) / pow;
        }

        /// <summary>
        /// 値の指定した桁の小数部を切り上げ丸める
        /// </summary>
        /// <param name="value">値</param>
        /// <param name="decimals">小数部桁数</param>
        /// <returns>丸めた値</returns>
        private static double RoundUp(double value, int decimals)
        {
            // 小数部桁数の10の累乗を取得
            double pow = Math.Pow(10, decimals);
            return 0 <= value ? Math.Ceiling(value * pow) / pow : Math.Floor(value * pow) / pow;
        }


        /// <summary>
        /// 長さの値を丸めた値取得
        /// </summary>
        /// <param name="value"></param>
        /// <param name="lengthRoundMode"></param>
        /// <param name="decimals"></param>
        /// <returns></returns>
        private double GetLengthRoundValue(double value, int lengthRoundMode, int decimals)
        {
            double roundValue = 0.0;

            // 切り捨ての場合
            if(lengthRoundMode == 0)
               roundValue = Truncate(value, decimals);
            // 切り上げの場合
            else if(lengthRoundMode == 1)
               roundValue = RoundUp(value, decimals);
            // 四捨五入の場合
            else if(lengthRoundMode == 2)
               roundValue = Math.Round(value, decimals, MidpointRounding.AwayFromZero);

            return roundValue;
        }

        /// <summary>
        /// 面積の値を丸めた値取得
        /// </summary>
        /// <param name="value"></param>
        /// <param name="areaRoundMode"></param>
        /// <param name="decimals"></param>
        /// <returns></returns>
        private double GetAreaRoundValue(double value, int areaRoundMode, int decimals)
        {
            double roundValue = 0.0;

            // 切り捨ての場合
            if(areaRoundMode == 0)
               roundValue = Truncate(value, decimals);
            // 切り上げの場合
            else if(areaRoundMode == 1)
               roundValue = RoundUp(value, decimals);
            // 四捨五入の場合
            else if(areaRoundMode == 2)
               roundValue = Math.Round(value, decimals, MidpointRounding.AwayFromZero);

            return roundValue;
        }

        /// <summary>
        /// 2点が同じ座標か
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <param name="scale"></param>
        /// <returns></returns>
        public Boolean IsPointEquals(Point2d point1, Point2d point2, double scale)
        {
            // 大きい座標空間で同一点を正確に判断できるように、レイアウトスケールをかける
            if(scale >= 0.0)
            {
                point1.X = point1.X * scale;
                point1.Y = point1.Y * scale;
                point2.X = point2.X * scale;
                point2.Y = point2.Y * scale;
            }

            return Geometry.IsPointEquals(point1, point2);
        }

        /// <summary>
        /// 2 線分の端点が一致するか
        /// </summary>
        /// <param name="line1"></param>
        /// <param name="line2"></param>
        /// <param name="scale"></param>
        /// <returns></returns>
        public Boolean IsLineEquals(in LineShape line1, in LineShape line2, double scale)
        {
            return (IsPointEquals(line1.StartPoint, line2.StartPoint, scale) &&
                IsPointEquals(line1.EndPoint, line2.EndPoint, scale)) ||
                (IsPointEquals(line1.StartPoint, line2.EndPoint, scale) &&
                IsPointEquals(line1.EndPoint, line2.StartPoint, scale));
        }

        /// <summary>
        /// 2点間の中心座標取得
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <returns></returns>
        public Point2d CenterPoint(Point2d point1, Point2d point2)
        {
            Point2d centerPoint = Geometry.GetMiddlePoint(point1, point2);
            return centerPoint;
        }


        /// <summary>
        /// 線分リストの線同士で交差しているものがあればそこで切って分けて新しいリストにする
        /// </summary>
        /// <param name="lineShapes"></param>
        /// <returns></returns>
        private List<SelectedShape> CutCrossLineShapes(List<SelectedShape> lineShapes)
        {
            List<SelectedShape> cutLineShapes = new List<SelectedShape>();

            int shapCount = lineShapes.Count;
            for(int i = 0; i < shapCount; i++)
            {
                LineShape line1 = (LineShape)lineShapes[i].Shape;
                Layout layout1 = lineShapes[i].Layout;
                List<double> intersectionPointOnShape1ParameterList = new List<double>();

                // 線分1の他の線分との交点パラメーターを取得
                for(int j = 0; j < shapCount; j++)
                {
                    if(i == j)
                        continue;

                    LineShape line2 = (LineShape)lineShapes[j].Shape;
                    IntersectionPoint[] intersectionPoints = line1.GetIntersectionPoint(line2);
                    int intersectionPointsCount = intersectionPoints.Length;
                    for(int k = 0; k < intersectionPointsCount; k++)
                    {
                        intersectionPointOnShape1ParameterList.Add(intersectionPoints[k].PointOnShape1.Parameter);
                    }
                }

                // 他の線分と交点（端点も含む）があった場合
                if(intersectionPointOnShape1ParameterList.Count >= 2)
                {
                    // 交点パラメータで切り分けた新しい線分をリストに追加
                    // パラメータの小さい順に並び替え
                    intersectionPointOnShape1ParameterList.Sort((a, b) => Math.Sign(a - b));
                    int intersectionPointOnShape1ParameterListCount = intersectionPointOnShape1ParameterList.Count;
                    for(int k = 0; k < intersectionPointOnShape1ParameterListCount - 1; k++)
                    {
                        double intersectionPointOnShape1Parameter1 = intersectionPointOnShape1ParameterList[k];
                        Point2d cutLineStartPoint = line1.GetPointAtParameter(intersectionPointOnShape1Parameter1);
                        double intersectionPointOnShape1Parameter2 = intersectionPointOnShape1ParameterList[k+1];
                        Point2d cutLineEndPoint = line1.GetPointAtParameter(intersectionPointOnShape1Parameter2);

                        if(!IsPointEquals(cutLineStartPoint, cutLineEndPoint, layout1.ScaleX))
                        {
                            // LineShape cutLine = (LineShape)line1.Copy(); // 注意 ドキュメント中にコピー図形が生成されるのでCopy()ではいけない
                            LineShape cutLine = ShapeFactory.CreateLine(cutLineStartPoint, cutLineEndPoint);

                            cutLine.SetPoint(cutLineStartPoint, cutLineEndPoint);

                            cutLineShapes.Add(new SelectedShape(cutLine, layout1));
                        }
                    }
                }
             }

            return cutLineShapes;
        }

        /// <summary>
        /// 三斜三角形の作成
        /// </summary>
        /// <param name="lineShapes"></param>
        /// <returns>※線同士が交差していない前提で処理するので、この関数の前でそのような線は交点で分けてリストに入れておく</returns>
        private List<TriclinicTriangle> CreateTriclinicTriangles(List<SelectedShape> lineShapes, int startNumber)
        {
            List<TriclinicTriangle> triclinicTriangles = new List<TriclinicTriangle>(); 

            int shapCount = lineShapes.Count;
            for(int i = 0; i < shapCount; i++)
            {
                LineShape line1 = (LineShape)lineShapes[i].Shape;
                Layout layout1 = lineShapes[i].Layout;
                Point2d point11 = line1.StartPoint;
                Point2d point12 = line1.EndPoint;

                // 1本目の線の終点と端点が一致する線を探す
                for(int j = i + 1; j < shapCount; j++)
                {
                    LineShape line2 = (LineShape)lineShapes[j].Shape;
                    Layout layout2 = lineShapes[j].Layout;

                    Point2d point21 = line2.StartPoint;
                    Point2d point22 = line2.EndPoint;

                    //同一点か
                    bool isEqualPoint21 = IsPointEquals(point12, point21, layout1.ScaleX);
                    bool isEqualPoint22 = IsPointEquals(point12, point22, layout1.ScaleX);

                    // 繋がる線が見つかる
                    if(isEqualPoint21 || isEqualPoint22)
                    {
                        Point2d checkPoint2 = isEqualPoint21 ? point22:point21;

                        // チェック用のLine2の無限線作成
                        InfiniteLineShape infiniteLine2Shape = ShapeFactory.CreateInfiniteLine(line2.StartPoint, line2.Angle);

                        // その線に繋がる線を探す
                        for(int k = i + 1; k < shapCount; k++)
                        {
                            if(k == j)
                                continue;

                            LineShape line3 = (LineShape)lineShapes[k].Shape;
                            Layout layout3 = lineShapes[k].Layout;

                            Point2d point31 = line3.StartPoint;
                            Point2d point32 = line3.EndPoint;

                            //同一点か
                            bool isEqualPoint31 = IsPointEquals(checkPoint2, point31, layout2.ScaleX);
                            bool isEqualPoint32 = IsPointEquals(checkPoint2, point32, layout2.ScaleX);

                            // 繋がる線が見つかる
                            if(isEqualPoint31 || isEqualPoint32)
                            {
                                Point2d checkPoint3 = isEqualPoint31 ? point32:point31;

                                // 最初の線の始点と繋がる場合、三角形ができる
                                if(IsPointEquals(checkPoint3, point11, layout3.ScaleX))
                                {
                                    int triclinicTrianglesCount = triclinicTriangles.Count;
                                    TriclinicTriangle triangle = new TriclinicTriangle(triclinicTrianglesCount + startNumber, point11, line2.StartPoint, line2.EndPoint, ShapeFactory, Geometry);
                                    // 有効な場合のみリストに追加
                                    if(triangle.baselineLength > RcPrecisionConfusion && triangle.hight > RcPrecisionConfusion)
                                        triclinicTriangles.Add(triangle);
                                }
                            }
                        }
                    }
                }
            }

            return triclinicTriangles;
        }

        /// <summary>
        /// 三斜三角形の高さや底辺長さを表す図形作成
        /// </summary>
        /// <param name="drawing"></param>
        /// <param name="triclinicTriangles"></param>
        /// <returns></returns>
        private List<Shape> CreateTriangularQuadratureShapes(Drawing drawing, List<TriclinicTriangle> triclinicTriangles)
        {
            Document doc = drawing.Document;

            // 書込みレイヤ
            LayerTable layerTable = doc.LayerTable;
            Layer currentLayer = layerTable.CurrentLayer;

             // 長さ数値の書式
            string lengthValueStringFormatParamf = "{0:f" + RC_LengthValueDecimals.ToString() + "}";// ex.{0:f2}
            
            // 文字高さ(三角形の高さ、底辺の長さ)
            double textHight = GetTextFontHeight(doc);

            // 番号の配置点
            Point2d[] balloonLocatePoint = new Point2d[1];

            // 図形リスト
            List<Shape> shapes = new List<Shape>();

            // 三角形ごとに高さや底辺の長さ番号の図形を作成
            foreach(TriclinicTriangle triclinicTriangle in triclinicTriangles)
            {
                // 高さを表す線 //
                Point2d vertexPoin = triclinicTriangle.vertexPoint;
                Point2d verticalPointOnBaseLine = triclinicTriangle.verticalPointOnBaseLine;
                LineShape heightLineShape = ShapeFactory.CreateLine(vertexPoin, verticalPointOnBaseLine);

                // 設定値
                doc.BasicShapeSettings.CopyToShape(heightLineShape);

                // 色
                if(RcAddinTriangularQuadratureHightLineColorNumber >= 0)
                    heightLineShape.ColorNumber = RcAddinTriangularQuadratureHightLineColorNumber;

                // 線種
                if(RcAddinTriangularQuadratureHightLineTypeNumber >= 0)
                    heightLineShape.LinetypeNumber = RcAddinTriangularQuadratureHightLineTypeNumber;

                // 線幅
                if(RcAddinTriangularQuadratureHightLineWidthNumber >= 0)
                    heightLineShape.LinewidthNumber = RcAddinTriangularQuadratureHightLineWidthNumber;

                // レイヤのセット
                heightLineShape.Layer = currentLayer;

                shapes.Add(heightLineShape);


                // 高さの文字 //
                double triangleHightValue = triclinicTriangle.hight;// 三角形の高さ
                double triangleHightValue2 = GetLengthRoundValue(triangleHightValue, RC_LengthRoundMode, RC_LengthValueDecimals);// 丸め
                string triangleHightText = String.Format(lengthValueStringFormatParamf, triangleHightValue2);

                // 配置点(高さの線の中点)（中下）
                Point2d textLocatePoint = CenterPoint(heightLineShape.StartPoint, heightLineShape.EndPoint);

                // 配置角度(向きを見やすく整えることはしない。隣の三角形の値と重なる場合があるため)
                double textAngle = heightLineShape.Angle;

                // 色
                int textColorNo = -1;
                if(RcAddinTriangularQuadratureTextColorNumber >= 0)
                    textColorNo = RcAddinTriangularQuadratureTextColorNumber;

                // 文字図形を作成してリストに追加
                AddTextShape(shapes, triangleHightText, textLocatePoint, textAngle, textHight, 0.0, Alignment.BottomMiddle, textColorNo, doc, currentLayer);

                // 底辺の長さの文字 //
                double baseLineLengthValue = triclinicTriangle.baselineLength;// 底辺の長さ
                double baseLineLengthValue2 = GetLengthRoundValue(baseLineLengthValue, RC_LengthRoundMode, RC_LengthValueDecimals);// 丸め
                string baseLineLengthText = String.Format(lengthValueStringFormatParamf, baseLineLengthValue2);

                // 配置点(底辺の線の中点)（中下）
                textLocatePoint = CenterPoint(triclinicTriangle.baseLineStartPoint, triclinicTriangle.baseLineEndPoint);

                // 配置角度(常に三角形の内側に表示させるようにする)
                textAngle = heightLineShape.Angle + 90.0;
                // 0度以上360度未満に補正
                textAngle = Geometry.WrapAngleTo360(textAngle);

                // 文字図形を作成してリストに追加
                AddTextShape(shapes, baseLineLengthText, textLocatePoint, textAngle, textHight, 0.0, Alignment.BottomMiddle, textColorNo, doc, currentLayer);

                // 三角形の番号 //
                balloonLocatePoint[0] = triclinicTriangle.gravityPoint;// 重心に置く
                string triangleNumbernumberText = triclinicTriangle.triangleNumber.ToString();
                BalloonShape balloonShape = ShapeFactory.CreateBalloon(triangleNumbernumberText, balloonLocatePoint);
                doc.BalloonSettings.CopyToShape(balloonShape);
                balloonShape.Size = GetBalloonSize(doc);

                // 番号文字の色
                if(textColorNo >= 0)
                {
                    balloonShape.UseTextColor = true;
                    balloonShape.TextColorNumber = textColorNo;
                }

                 // レイヤのセット
                balloonShape.Layer = currentLayer;

                shapes.Add(balloonShape);
            }

            return shapes;
        }

        /// <summary>
        /// 三斜求積表の図形作成
        /// </summary>
        /// <param name="drawing"></param>
        /// <param name="triclinicTriangles"></param>
        /// <param name="locatePoint">表の配置点(左上)座標</param>
        /// <returns></returns>
        private List<Shape> CreateTriangularQuadratureTableShapes(Drawing drawing, List<TriclinicTriangle> triclinicTriangles, Point2d locatePoint)
        {
            Document doc = drawing.Document;

            // 書込みレイヤ
            LayerTable layerTable = doc.LayerTable;
            Layer currentLayer = layerTable.CurrentLayer;
            
            // 文字高さ(三角形の高さ、底辺の長さ)
            double textHight = GetTextFontHeight(doc);

            // データー数
            int listCount = triclinicTriangles.Count;

            // セル内での文字オフセット
            double textOffsetH = textHight * 0.25;
            double textOffsetW = textHight * 0.5;

            // 表のサイズ
            double cellHight = textHight + textOffsetH * 2.0;// セル高さ
            int headRowsCount = 1; // 項目行などデータ前の行数
            int tailRowsCount = 3; // 合計値(倍面積合計、面積、敷地面積)などのデータの後の行数
            double tableHight = cellHight * (double)(listCount + headRowsCount + tailRowsCount);// 表の高さ、項目行と合計値の行分多くする
            double tableWidth = textHight * RcTriangularQuadratureTableWidthTextScale;// 表の幅
            string[] clumnNames = new string[] {"番号", "底 辺", "高 さ", "倍 面 積"};
            int clumnCount = clumnNames.Length;// (列の数) 
            int[] widthRatio = new int[] {8, 20, 20, 30}; // 各セル幅の比率
            Debug.Assert(clumnCount == widthRatio.Length);
            int totalWidthRatio = 0;
            foreach(int ratio in widthRatio)
                totalWidthRatio += ratio;

            // 各セルの幅
            double[] cellWidth = new double[clumnCount];
            for(int i = 0; i < clumnCount; i++)
                cellWidth[i] = tableWidth * ((double)widthRatio[i] / (double)totalWidthRatio);

            string[] totalValueNames = new string[] { "倍面積合計", "面      積", "敷 地 面 積" };

            // 図形リスト
            List<Shape> shapes = new List<Shape>();

            // 表の配置点(左上)
            Point2d tableLocatePointTL = locatePoint;

            // 色
            int textColorNo = -1;
            if(RcAddinTriangularQuadratureTableTextColorNumber >= 0)
                textColorNo = RcAddinTriangularQuadratureTableTextColorNumber;

            
            double totalWidth = 0.0;
            double totalAreaW = 0.0; // 倍面積合計
            string valueStringFormatParamf = "";
            double dataValue = 0.0;
            string dataValueText = "";

            // 縦線
            for(int i = 0; i < clumnCount + 1; i++)
            {
                double lineLength = tableHight;
                if(i > 0 && i < 2)
                    lineLength -= cellHight * (double)totalValueNames.Length;// "倍面積合計", "面積", "敷地面積" 行分

                if(i > 0)
                    totalWidth += cellWidth[i-1];

                Point2d lineStartPoint = new Point2d(tableLocatePointTL.X + totalWidth, tableLocatePointTL.Y);
                Point2d lineEndPoint = new Point2d(lineStartPoint.X, tableLocatePointTL.Y - lineLength);

                LineShape lineShape = ShapeFactory.CreateLine(lineStartPoint, lineEndPoint);
                doc.BasicShapeSettings.CopyToShape(lineShape);

                // 図形追加
                // レイヤのセット
                lineShape.Layer = currentLayer;

                // 色
                if(RcAddinTriangularQuadratureTableLineColorNumber >= 0)
                    lineShape.ColorNumber = RcAddinTriangularQuadratureTableLineColorNumber;

                // 線種
                lineShape.LinetypeNumber = 1;

                // 線幅
                if(RcAddinTriangularQuadratureTableLineLineWidthNumber >= 0)
                    lineShape.LinewidthNumber = RcAddinTriangularQuadratureTableLineLineWidthNumber;

                // 追加
                shapes.Add(lineShape);
            }

            // 横線
            for(int i = 0; i < headRowsCount + listCount + tailRowsCount + 1; i++)
            {
                double lineLength = tableWidth;
 
                Point2d lineStartPoint = new Point2d(tableLocatePointTL.X, tableLocatePointTL.Y - cellHight * (double)i);
                Point2d lineEndPoint = new Point2d(tableLocatePointTL.X + lineLength, lineStartPoint.Y);

                LineShape lineShape = ShapeFactory.CreateLine(lineStartPoint, lineEndPoint);
                doc.BasicShapeSettings.CopyToShape(lineShape);

                // 図形追加
                // レイヤのセット
                lineShape.Layer = currentLayer;

                // 色
                if(RcAddinTriangularQuadratureTableLineColorNumber >= 0)
                    lineShape.ColorNumber = RcAddinTriangularQuadratureTableLineColorNumber;

                // 線種
                lineShape.LinetypeNumber = 1;

                // 線幅
                if(RcAddinTriangularQuadratureTableLineLineWidthNumber >= 0)
                    lineShape.LinewidthNumber = RcAddinTriangularQuadratureTableLineLineWidthNumber;

                // 追加
                shapes.Add(lineShape);
            }

            // 文字
            for(int i = 0; i < headRowsCount + listCount + tailRowsCount; i++)
            {
                totalWidth = 0.0;

                // 項目名
                if(i == 0)
                {
                    for(int j = 0; j < clumnCount; j++)
                    {
                        totalWidth += cellWidth[j];

                        // 配置点（中中）
                        Point2d textLocatePoint = new Point2d(tableLocatePointTL.X + totalWidth - cellWidth[j] * 0.5, tableLocatePointTL.Y - cellHight * 0.5 - cellHight * (double)i);

                        // 文字図形を作成してリストに追加
                        AddTextShape(shapes, clumnNames[j], textLocatePoint, 0.0, textHight, 0.0, Alignment.MiddleMiddle, textColorNo, doc, currentLayer);
                    }
                }
                // データの下の合計値等
                else if( i >= headRowsCount + listCount)
                {
                    string dataName = "";
                    dataValueText = " ";

                    // 倍面積合計
                    if(i == headRowsCount + listCount)
                    {
                        dataName = totalValueNames[0];
                        dataValue = GetLengthRoundValue(totalAreaW, RC_AreaRoundMode, RC_AreaWValueDecimals);// 丸め
                        valueStringFormatParamf = "{0:f" + RC_AreaWValueDecimals.ToString() + "}";// 倍面積数値の書式 ex.{0:f4}
                        dataValueText = String.Format(valueStringFormatParamf, dataValue);
                    }
                    // 面積
                    else if(i == headRowsCount + listCount + 1)
                    {
                        dataName = totalValueNames[1];
                        dataValue = totalAreaW * 0.5;
                        dataValue = GetLengthRoundValue(dataValue, RC_AreaRoundMode, RC_AreaValueDecimals);// 丸め
                        valueStringFormatParamf = "{0:f" + RC_AreaValueDecimals.ToString() + "}";// 面積数値の書式 ex.{0:f5}
                        dataValueText = String.Format(valueStringFormatParamf, dataValue);
                    }
                    // 敷地面積
                    else if(i == headRowsCount + listCount + 2)
                    {
                        dataName = totalValueNames[2];
                        dataValue = totalAreaW * 0.5;
                        dataValue = GetLengthRoundValue(dataValue, RC_AreaRoundMode, RC_SiteAreaValueDecimals);// 丸め
                        valueStringFormatParamf = "{0:f" + RC_SiteAreaValueDecimals.ToString() + "}";// 敷地面積数値の書式 ex.{0:f2}
                        dataValueText = String.Format(valueStringFormatParamf, dataValue) + " ㎡";// 単位付加
                    }

                    // 項目名
                    totalWidth = cellWidth[0] + cellWidth[1];

                    // 配置点(中中)
                    Point2d textLocatePoint = new Point2d(tableLocatePointTL.X + totalWidth + cellWidth[2] * 0.5, tableLocatePointTL.Y - cellHight * 0.5 - cellHight * (double)i);

                    // 文字図形を作成してリストに追加
                    AddTextShape(shapes, dataName, textLocatePoint, 0.0, textHight, 0.0, Alignment.MiddleMiddle, textColorNo, doc, currentLayer);

                    // 値
                    totalWidth += cellWidth[2] + cellWidth[3];

                    // 配置点(右下)
                    Point2d textLocatePoint2 = new Point2d(tableLocatePointTL.X + totalWidth - textOffsetW, tableLocatePointTL.Y - cellHight * (double)(i + 1) + textOffsetH);

                    // 文字図形を作成してリストに追加
                    AddTextShape(shapes, dataValueText, textLocatePoint2, 0.0, textHight, 0.0, Alignment.BottomRight, textColorNo, doc, currentLayer);
                }
                // データ部分
                else
                {
                    // 三角形
                    TriclinicTriangle triclinicTriangle = triclinicTriangles[i - 1];

                    string baselineLengthText = "";
                    string hightText = "";

                    for(int j = 0; j < clumnCount; j++)
                    {
                        totalWidth += cellWidth[j];

                        switch(j)
                        {
                            // 番号
                            case 0:
                                dataValueText = triclinicTriangle.triangleNumber.ToString();
                                break;
                            // 底辺
                            case 1:
                                dataValue = triclinicTriangle.baselineLength;
                                dataValue = GetLengthRoundValue(dataValue, RC_LengthRoundMode, RC_LengthValueDecimals);// 丸め
                                valueStringFormatParamf = "{0:f" + RC_LengthValueDecimals.ToString() + "}";// // 長さ数値の書式 ex.{0:f2}
                                dataValueText = String.Format(valueStringFormatParamf, dataValue);
                                baselineLengthText = dataValueText;
                                break;
                            // 高さ
                            case 2:
                                dataValue = triclinicTriangle.hight;
                                dataValue = GetLengthRoundValue(dataValue, RC_LengthRoundMode, RC_LengthValueDecimals);// 丸め
                                valueStringFormatParamf = "{0:f" + RC_LengthValueDecimals.ToString() + "}";// // 長さ数値の書式 ex.{0:f2}
                                dataValueText = String.Format(valueStringFormatParamf, dataValue);
                                hightText = dataValueText;
                                break;
                            // 倍面積
                            case 3:
                                Debug.Assert(baselineLengthText != "" && hightText != "");

                                // 文字列をdouble型に変換
                                double.TryParse(baselineLengthText, out double baselineLengthTextValue);
                                double.TryParse(hightText, out double hightTextValue);
                                // dataValue = triclinicTriangle.triangleAreaW;// 2021/01/04 変更:表の底辺と高さの値を使った計算結果にする（表を見て手動計算した結果と同じになるように）
                                dataValue = baselineLengthTextValue * hightTextValue;
                                dataValue = GetLengthRoundValue(dataValue, RC_AreaRoundMode, RC_AreaValueDecimals - 1);// 丸め
                                totalAreaW += dataValue;

                                valueStringFormatParamf = "{0:f" + (RC_AreaValueDecimals - 1).ToString() + "}";// // 面積数値の書式 ex.{0:f4}
                                dataValueText = String.Format(valueStringFormatParamf, dataValue);
                                break;
                        }

                        // 配置点(右下)
                        Point2d textLocatePoint = new Point2d(tableLocatePointTL.X + totalWidth - textOffsetW, tableLocatePointTL.Y - cellHight * (double)(i + 1) + textOffsetH);

                        // 文字図形を作成してリストに追加
                        AddTextShape(shapes, dataValueText, textLocatePoint, 0.0, textHight, 0.0, Alignment.BottomRight, textColorNo, doc, currentLayer);
                    }
                }
            }

            return shapes;
        }


        /// <summary>
        /// 図形リストに指定した設定の文字の文字図形を追加
        /// </summary>
        /// <param name="shapes"></param>
        /// <param name="str"></param>
        /// <param name="locatePoint"></param>
        /// <param name="angle"></param>
        /// <param name="textHight"></param>
        /// <param name="lineSpacing"></param>
        /// <param name="alignment"></param>
        /// <param name="colorNumber"></param>
        /// <param name="doc"></param>
        /// <param name="currentLayer"></param>
        private TextShape AddTextShape(List<Shape> shapes, string str, Point2d locatePoint, double angle, double textHight, double lineSpacing, Alignment alignment, int colorNumber, Document doc, Layer currentLayer)
        {
            TextShape textShape = ShapeFactory.CreateText(str, locatePoint, angle);
            doc.TextSettings.CopyToShape(textShape);

            // 文字高さセット
            textShape.FontHeight = textHight;

            // 向きは横書き固定
            textShape.DirectionVertical = false;

            // 行間をセット
            textShape.LineSpacing = lineSpacing;

            // 配置基準位置
            textShape.Alignment = alignment;

            // 色
            if(colorNumber >= 0)
                textShape.ColorNumber = colorNumber;

            // レイヤのセット
            textShape.Layer = currentLayer;

            // リストに図形の追加
            shapes.Add(textShape);

            return textShape;
        }

        /// <summary>
        /// 現在のドキュメントの文字高さを取得する
        /// 用紙を表示している場合は部分図レイアウトの逆スケール倍した値
        /// </summary>
        /// <param name="doc"></param>
        /// <returns></returns>
        private　double GetTextFontHeight(Document doc)
        {
            double textFontHeight = doc.TextSettings.FontHeight;

            // 用紙を表示している場合は部分図レイアウトの逆スケール倍した値
            if(GetSizeScale(doc, out double scaleX, out double scaleY))
                textFontHeight *= scaleX;

            return textFontHeight;
        }

        /// <summary>
        /// 現在のドキュメントのバルーンサイズを取得する
        /// 用紙を表示している場合は部分図レイアウトの逆スケール倍した値
        /// </summary>
        /// <param name="doc"></param>
        /// <returns></returns>
        private　double GetBalloonSize(Document doc)
        {
            double balloonSize = doc.BalloonSettings.Size;

            // 用紙を表示している場合は部分図レイアウトの逆スケール倍した値
            if(GetSizeScale(doc, out double scaleX, out double scaleY))
                balloonSize *= scaleX;

            return balloonSize;
        }

        /// <summary>
        /// 大きさに適用するスケールを取得
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="scaleX"></param>
        /// <param name="scaleY"></param>
        /// <returns></returns>
        private　Boolean GetSizeScale(Document doc, out double scaleX, out double scaleY)
        {
            // 結果の初期化
            scaleX = 1.0;
            scaleY = 1.0;

            // 現在のレイアウト
            Layout layout = doc.CurrentLayout;
            if(layout == null)
                return false;

            if(layout.IsCurrent)
            {
                double layoutScaleX = layout.ScaleX;
                double layoutScaleY = layout.ScaleY;

                if(layoutScaleX <= RcPrecisionScale || layoutScaleY <= RcPrecisionScale)
                    return false;

                scaleX = 1.0 / layoutScaleX;
                scaleY = 1.0 / layoutScaleY;

                return true;
            }

            return true;
        }

    }    
}
